package java.util;

import java.lang.invoke.SerializedLambda;
import java.util.function.ToDoubleFunction;
import java.util.function.ToLongFunction;
import java.util.function.ToIntFunction;
import java.util.function.Function;
import java.io.Serializable;

@FunctionalInterface
public interface Comparator<T>
{
    public abstract int compare(T p0, T p1);
    
    public abstract boolean equals(Object p0);
    
    default public Comparator<T> reversed() {
        return Collections.reverseOrder(this);
    }
    
    default public Comparator<T> thenComparing(Comparator<? super T> comparator) {
        Objects.requireNonNull(comparator);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> {
            int n = this.compare(o, o2);
            return (n != 0) ? n : p0.compare(o, o2);
        });
    }
    
    default public <U> Comparator<T> thenComparing(Function<? super T, ? extends U> function, Comparator<? super U> comparator) {
        return this.thenComparing(comparing(function, comparator));
    }
    
    default public <U extends Comparable<? super U>> Comparator<T> thenComparing(Function<? super T, ? extends U> function) {
        return this.thenComparing(comparing(function));
    }
    
    default public Comparator<T> thenComparingInt(ToIntFunction<? super T> toIntFunction) {
        return this.thenComparing(comparingInt(toIntFunction));
    }
    
    default public Comparator<T> thenComparingLong(ToLongFunction<? super T> toLongFunction) {
        return this.thenComparing(comparingLong(toLongFunction));
    }
    
    default public Comparator<T> thenComparingDouble(ToDoubleFunction<? super T> toDoubleFunction) {
        return this.thenComparing(comparingDouble(toDoubleFunction));
    }
    
    public static <T extends Comparable<? super T>> Comparator<T> reverseOrder() {
        return Collections.reverseOrder();
    }
    
    public static <T extends Comparable<? super T>> Comparator<T> naturalOrder() {
        return Comparators.NaturalOrderComparator.INSTANCE;
    }
    
    public static <T> Comparator<T> nullsFirst(Comparator<? super T> comparator) {
        return new Comparators.NullComparator(true, comparator);
    }
    
    public static <T> Comparator<T> nullsLast(Comparator<? super T> comparator) {
        return new Comparators.NullComparator(false, comparator);
    }
    
    public static <T, U> Comparator<T> comparing(Function<? super T, ? extends U> function, Comparator<? super U> comparator) {
        Objects.requireNonNull(function);
        Objects.requireNonNull(comparator);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> p1.compare(p0.apply(o), p0.apply(o2)));
    }
    
    public static <T, U extends Comparable<? super U>> Comparator<T> comparing(Function<? super T, ? extends U> function) {
        Objects.requireNonNull(function);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> ((Comparable)p0.apply(o)).compareTo(p0.apply(o2)));
    }
    
    public static <T> Comparator<T> comparingInt(ToIntFunction<? super T> toIntFunction) {
        Objects.requireNonNull(toIntFunction);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> Integer.compare(p0.applyAsInt(o), p0.applyAsInt(o2)));
    }
    
    public static <T> Comparator<T> comparingLong(ToLongFunction<? super T> toLongFunction) {
        Objects.requireNonNull(toLongFunction);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> Long.compare(p0.applyAsLong(o), p0.applyAsLong(o2)));
    }
    
    public static <T> Comparator<T> comparingDouble(ToDoubleFunction<? super T> toDoubleFunction) {
        Objects.requireNonNull(toDoubleFunction);
        return (Comparator)(Serializable)((o, o2)/* java.lang.Object,java.lang.Object, */ -> Double.compare(p0.applyAsDouble(o), p0.applyAsDouble(o2)));
    }
}
